"""Run slideshow of logos.
"""

from lpminimk3 import Mode, find_launchpads
from lpminimk3.graphics import Bitmap
from lpminimk3.graphics.art import Bitmaps
import time
import os


SLIDES = [
        "APPLE",
        "APPLE_OLD",
        "FACEBOOK",
        "GMAIL",
        "GOOGLE",
        "ITUNES",
        "WINDOWS",
        "XBOX",
        "YOUTUBE"
        ]


def main():
    lp = find_launchpads()[0]  # Get the first available launchpad
    lp.open()  # Open device for reading and writing on MIDI interface (by default)  # noqa

    lp.mode = Mode.PROG  # Switch to the programmer mode

    print('Watch logos appear on your Launchpad.\n'  # noqa
          'Press Ctrl+C to quit.\n')

    while True:
        try:
            for k, v in Bitmaps.__dict__.items():
                if k in SLIDES and isinstance(v, str) and os.path.isfile(v):
                    lp.grid.render(Bitmap(v))
                    time.sleep(2)
        except KeyboardInterrupt:
            print('\n')
            break


if __name__ == '__main__':
    main()



"""High-level rendering API for the Launchpad Mini MK3's surface.

This module provides a high-level abstraction for rendering
on the Launchpad Mini MK3.
"""

from ._graphics import (Text,  # noqa
                        Bitmap,
                        Movie,
                        Frame,
                        Renderable,
                        ScrollDirection,
                        FlipAxis)



"""Web color.

This module is an abstraction over web colors. The full list
of colors can be found on
`Wikipedia <https://en.wikipedia.org/wiki/Lists_of_colors>`_.

Examples
--------

Set color of LED "0x0" to "amethyst":
    >>> import lpminimk3
    >>> from lpminimk3.colors import WebColor
    >>> lp = lpminimk3.find_launchpads()[0]
    >>> lp.open()
    >>> lp.mode = 'prog'
    >>> lp.grid.led(0, 0).color = WebColor('amethyst')

Print hex, RGB and name for color:
    >>> from lpminimk3.colors import WebColor
    >>> WebColor('magenta').hex
    '#f0f'
    >>> WebColor('magenta').rgb
    (255, 0, 255)
    >>> WebColor('magenta').name
    'magenta'

Check if color exists in color dictionary:
    >>> from lpminimk3.colors import WebColor
    >>> 'yellow' in WebColor.COLOR_DICT
    True
    >>> 'brown' in WebColor.COLOR_DICT
    False
    >>> 'teal' in WebColor.COLOR_DICT
    True
    >>> 'magenta' in WebColor.COLOR_DICT
    True

Print full details for color in color dictionary:
    >>> from lpminimk3.colors import WebColor
    >>> WebColor.COLOR_DICT['crimson']
    {'name': 'Crimson', 'hex': '#dc143c', 'rgb': [220, 20, 60]}
"""


import os
import json
import jsonschema


class WebColorDictionary:
    """Web color dictionary.
    """
    def __init__(self,
                 json_filename, *,
                 schema_filename='./web_colors.schema.json'):
        """Constructs a web color dictionary from JSON file
        `json_filename`.

        Parameters
        ----------
        json_filename : str
            Name of JSON file to parse.
        schema_filename : str, optional
            Name of schema JSON file.
        """
        self._filename = self._determine_abspath(json_filename)
        self._data = self._load(json_filename)
        schema = self._load(schema_filename)
        self._validate(self._data, schema)

    def __iter__(self):
        for color, properties in self._data.items():
            yield {color: properties}

    def __contains__(self, color):
        return color in self._data

    def __getitem__(self, color):
        return self._data[color]

    def __repr__(self):
        return ("WebColorDictionary("
                f"filename='{os.path.basename(self.filename)}')")

    def __str__(self):
        return (str(self._data)
                if self._data
                else '')

    @property
    def filename(self):
        """File name.
        """
        return self._filename

    @property
    def data(self):
        """Data.
        """
        return self._data

    def _load(self, filename):
        data = None
        filename = self._determine_abspath(filename)
        with open(filename) as f:
            data = json.load(f)
        return data

    def _validate(self, data, schema):
        try:
            jsonschema.validate(instance=data, schema=schema)
        except jsonschema.exceptions.ValidationError:
            raise ValueError('Invalid JSON file format.')

    def _determine_abspath(self, filename):
        if not os.path.isabs(filename):
            current_dir = os.path.dirname(__file__)
            return os.path.join(current_dir, filename)
        return filename


class WebColor:
    """Web color.

    Example
    -------
    Set color of LED "0x0" to "amethyst":
        >>> import lpminimk3
        >>> from lpminimk3.colors import WebColor
        >>> lp = lpminimk3.find_launchpads()[0]
        >>> lp.open()
        >>> lp.mode = 'prog'
        >>> lp.grid.led(0, 0).color = WebColor('amethyst')
    """

    """Web color dictionary.
    """
    COLOR_DICT = WebColorDictionary('./web_colors.json')

    def __init__(self, name):
        """Constructs a web color with name `name`.

        Parameters
        ----------
        name : str
            Name of web color.
        """
        self._name = name

    @property
    def name(self):
        """Name.
        """
        return self._name

    @property
    def rgb(self):
        """RGB value.
        Its possible values range from 0 to 255.
        """
        return tuple(WebColor.COLOR_DICT[self._name]['rgb'])

    @property
    def rgb_normalized(self):
        """RGB value, normalized to work on a Launchpad.
        Its possible values range from 0 to 127.
        """
        r, g, b = WebColor.COLOR_DICT[self._name]['rgb']
        return (r >> 1, g >> 1, b >> 1)

    @property
    def hex(self):
        """Hex color.
        """
        return WebColor.COLOR_DICT[self._name]['hex']

    def __repr__(self):
        return ("WebColor("
                f"name='{self.name}')")


if __name__ == "__main__":
    import doctest
    doctest.testmod()



"""Scroll the text "Hello, world!" from right to left
across the Launchpad's surface once.
"""

from lpminimk3 import Mode, find_launchpads
from lpminimk3.graphics import Text


def main():
    """Scroll text from right to left across the Launchpad's surface.
    """
    lp = find_launchpads()[0]  # Get the first available launchpad
    lp.open()  # Open device for reading and writing on MIDI interface (by default)  # noqa

    lp.mode = Mode.PROG  # Switch to the programmer mode

    lp.grid.render(Text(' Hello, world!').scroll())  # Scroll text once  # noqa


if __name__ == '__main__':
    main()



"""Python API for the Launchpad Mini MK3.

This module exposes API to discover connected Launchpads
on your computer and control them.

Examples
--------

Find a connected device and open the device for reading and writing:
    >>> import lpminimk3
    >>> lp = lpminimk3.find_launchpads()[0]
    >>> lp.open()

Query the device to ensure reading and writing works:
    >>> lp.device_inquiry()
    MidiEvent(message=[240, 0, 32, 41, 2, 13, 14, 1, 247], deltatime=0.0)

Switch to the programmer mode:
    >>> lp.mode = Mode.PROG

After a Launchpad is initialized, its `Grid` and `Panel` can be manipulated.

Set color of LED at "0x0" to yellow (Valid values: 0 - 127):
    >>> lp.grid.led('0x0').color = 10

Set color of LED at "1x0" to the first shade of palette color "red":
    >>> import lpminimk3
    >>> from lpminimk3.colors import ColorPalette
    >>> lp.grid.led('1x0').color = ColorPalette.Red.SHADE_1

Set color of LED at "logo" (in panel) to violet:
    >>> lp.panel.led('logo').color = 'violet'

Set color of LED at "Drums" to second shade of "green":
    >>> lp.panel.led('drums').color = 'green2'

Set color of LED at "Stop/Solo/Mute" to first shade of "white":
    >>> lp.panel.led('stop').color = 'w1'

Set color of LED at "Stop/Solo/Mute" to third shade of "orange":
    >>> lp.panel.led('mute').color = 'o3'

Set color of LED at "Stop/Solo/Mute" to first shade of "red" (Invalid, but okay):  # noqa
    >>> lp.panel.led('solo').color = 'r0'

Set color of LED at "Scene Launch 1" to "red" using hex:
    >>> lp.panel.led('scene_launch_1').color = '#ff0000'

Set color of LED at "Scene Launch 2" to "blue" using RGB:
    >>> lp.panel.led('scene_launch_2').color = (0, 0, 255)

Set color of LED at "Scene Launch 2" to "magenta" using web colors:
    >>> from lpminimk3.colors import WebColor
    >>> lp.panel.led('scene_launch_3').color = WebColor("magenta")

Turn off "Stop/Solo/Mute" LED:
    >>> lp.panel.led('mute').color = 0

Another way to turn off LED:
    >>> lp.panel.led('mute').reset()

Another way to turn off LED:
    >>> del lp.panel.led('mute').color

Wait for and respond to button presses and releases:
    >>> lp.panel.buttons().poll_for_event()
    ButtonEvent(button='4x1', type='press', deltatime=4.202155996)

Wait for button releases:
    >>> lp.panel.buttons().poll_for_event(type='release')
    ButtonEvent(button='4x1', type='release', deltatime=0.171533228)

Wait for "Up", "0x0" or "Stop/Solo/Mute" button to be pressed or released:
    >>> lp.panel.buttons('up', '0x0', 'stop').poll_for_event()
    ButtonEvent(button='4x1', type='release', deltatime=0.040492674)
"""

from .__version__ import __version__, VERSION  # noqa
from .components import Grid, Panel, Led, ButtonFace  # noqa
from .utils import Interface, Mode, Layout, ButtonEvent  # noqa
from .device import LaunchpadMiniMk3, find_launchpads  # noqa

if __name__ == "__main__":
    import doctest
    doctest.testmod()



import os

_CURRENT_DIR = os.path.dirname(__file__)


class Bitmaps:
    APPLE = os.path.join(_CURRENT_DIR, 'bitmaps/apple.bitmap.json')  # noqa
    APPLE_OLD = os.path.join(_CURRENT_DIR, 'bitmaps/apple_old.bitmap.json')  # noqa
    FACEBOOK = os.path.join(_CURRENT_DIR, 'bitmaps/facebook.bitmap.json')  # noqa
    GMAIL = os.path.join(_CURRENT_DIR, 'bitmaps/gmail.bitmap.json')  # noqa
    GOOGLE = os.path.join(_CURRENT_DIR, 'bitmaps/google.bitmap.json')  # noqa
    ITUNES = os.path.join(_CURRENT_DIR, 'bitmaps/itunes.bitmap.json')  # noqa
    PLUG = os.path.join(_CURRENT_DIR, 'bitmaps/plug.bitmap.json')  # noqa
    SMILEY = os.path.join(_CURRENT_DIR, 'bitmaps/smiley.bitmap.json')  # noqa
    WINDOWS = os.path.join(_CURRENT_DIR, 'bitmaps/windows.bitmap.json')  # noqa
    XBOX = os.path.join(_CURRENT_DIR, 'bitmaps/xbox.bitmap.json')  # noqa
    YOUTUBE = os.path.join(_CURRENT_DIR, 'bitmaps/youtube.bitmap.json')  # noqa


class Movies:
    PING_PONG = os.path.join(_CURRENT_DIR, 'movies/ping_pong.movie.json')  # noqa



"""Utility classes for Launchpad Mini MK3.

This module contains classes that represent the logical
parts of the Launchpad i.e. parts that have not physical
representation (like components).
"""

import enum
import time
import re
import platform
from collections import namedtuple
from . import _logging
from .match import Match

logger = _logging.getLogger(__name__)

MIDI_MESSAGE_LENGTH = 9


class MidiEvent:
    """A MIDI event.

    A MIDI event is received every time the Launchpad's MIDI port
    is read.
    """

    def __init__(self, message, deltatime=0):
        self._message = message
        self._deltatime = deltatime

    def __eq__(self, other):
        if not isinstance(other, (MidiEvent, list)):
            return False
        if isinstance(other, list):
            return self.message == other
        return self.message == other.message

    @property
    def message(self):
        """Message.
        """
        return self._message

    @property
    def deltatime(self):
        """Deltatime.
        """
        return self._deltatime

    def __repr__(self):
        return ('MidiEvent('
                f'message={self.message}, '
                f'deltatime={self.deltatime})')


class ButtonEvent:
    """A button event.

    A button event is received every time a button on the Launchpad
    is pushed.
    """

    PRESS = 'press'
    RELEASE = 'release'
    PRESS_RELEASE = 'press_release'

    def __init__(self, midi_event, buttons):
        self._midi_event = midi_event
        self._button = self._determine_button(midi_event, buttons)

    def __eq__(self, other):
        if not isinstance(other, ButtonEvent):
            return False
        return self.midi_event == other.midi_event

    def __repr__(self):
        return ("ButtonEvent("
                f"button='{self.button.name}', "
                f"type='{self.type}', "
                f"deltatime={self.deltatime})")

    @property
    def message(self):
        """Message.
        """
        return self._midi_event.message

    @property
    def deltatime(self):
        """Deltatime.
        """
        return self._midi_event.deltatime

    @property
    def button(self):
        """Button.
        """
        return self._button

    @property
    def type(self):
        """Type.
        """
        if not self._midi_event:
            return ''
        return (ButtonEvent.RELEASE
                if self._midi_event
                and len(self._midi_event.message) == 3
                and self._midi_event.message[2] == 0x0
                else ButtonEvent.PRESS)

    def _determine_button(self, midi_event, buttons):
        found_buttons = (list(filter(lambda b: b.midi_value == midi_event.message[1], buttons))  # noqa
                         if midi_event and midi_event.message and len(midi_event.message) == 3  # noqa
                         else None)
        return (found_buttons[0]
                if found_buttons and len(found_buttons)
                else None)


class MidiPort:
    """A MIDI port.
    """

    OUT = 'out'
    IN = 'in'
    DEFAULT_CLIENT_NAME = 'lpminimk3'

    def __init__(self, port_name, port_number, port_index,
                 system_port_name, *, direction,
                 midi_in=None, midi_out=None,
                 virtual=False):
        self._port_name = port_name
        self._port_number = port_number
        self._port_index = port_index
        self._system_port_name = system_port_name
        self._midi_in = midi_in
        self._midi_out = midi_out
        self._direction = direction
        self._virtual = virtual
        if midi_in:
            midi_in.ignore_types(sysex=False, timing=False)

    def __eq__(self, other):
        if not isinstance(other, MidiPort):
            return False
        return self.system_port_name == other.system_port_name

    def __repr__(self):
        return ("MidiPort("
                f"name='{self.port_name}', "
                f"number={self.port_number}, "
                f"index={self.port_index})")

    def __exit__(self, *args, **kwargs):
        self.close()

    @property
    def port_name(self):
        """Port name.
        """
        return self._port_name

    @property
    def port_number(self):
        """Port number.
        """
        return self._port_number

    @property
    def port_index(self):
        """Port index.
        """
        return self._port_index

    @property
    def system_port_name(self):
        """System port name.
        """
        return self._system_port_name

    @property
    def midi_in_handle(self):
        """MIDI IN handle.
        """
        return self._midi_in

    @property
    def midi_out_handle(self):
        """MIDI OUT handle.
        """
        return self._midi_out

    def is_open(self):
        """Checks to see if port is open.

        Returns
        -------
        bool
            `True` if port is open, otherwise `False`.
        """
        if self._midi_in and self._direction == MidiPort.IN:
            return self._midi_in.is_port_open()
        elif self._midi_out and self._direction == MidiPort.OUT:
            return self._midi_out.is_port_open()
        return False

    def open(self):
        """Opens MIDI port.
        """
        if (self._direction == MidiPort.OUT
                and not self._midi_out.is_port_open()):
            if self._virtual:
                self._midi_out.open_virtual_port(self._system_port_name)
            else:
                self._midi_out.open_port(self.port_index, MidiPort.OUT)

            if platform.system() != 'Windows' and platform.system() != 'Darwin':  # noqa
                self._midi_out.set_client_name(MidiPort.DEFAULT_CLIENT_NAME)

        elif (self._direction == MidiPort.IN
                and not self._midi_in.is_port_open()):
            if self._virtual:
                self._midi_in.open_virtual_port(self._system_port_name)
            else:
                self._midi_in.open_port(self.port_index, MidiPort.IN)

            if platform.system() != 'Windows' and platform.system() != 'Darwin':  # noqa
                self._midi_in.set_client_name(MidiPort.DEFAULT_CLIENT_NAME)

    def close(self):
        """Closes MIDI port.
        """
        if self.is_open():
            if self._midi_out and self._direction == MidiPort.OUT:
                self._midi_out.close_port()
            elif self._midi_in and self._direction == MidiPort.IN:
                self._midi_in.close_port()

    def send_message(self, message):
        """Sends raw MIDI message to Launchpad.

        Parameters
        ----------
        message : list of str
            Message of raw integers to send to Launchpad.
        """
        if (not message
                or (not isinstance(message, list)
                    and not hasattr(message, 'data'))):
            raise TypeError('Message must be of type list or MidiMessage.')
        message = (message.data
                   if hasattr(message, 'data')
                   else message)

        if self._midi_in and self._direction == MidiPort.IN:
            self._midi_in.send_message(message)
            logger.debug(f'MIDI message sent: {message}')
        elif self._midi_out and self._direction == MidiPort.OUT:
            self._midi_out.send_message(message)
            logger.debug(f'MIDI message sent: {message}')
        else:
            raise RuntimeError('Failed to send message.')

    def poll_for_event(self, *, timeout=5, match=None, read_delay=.001):
        """Polls for button event.

        Parameters
        ----------
        timeout : int
            Timeout in seconds.
        match : Match or None
            Button match.
        read_delay : float
            Delay duration between reads.

        Returns
        -------
        ButtonEvent or None
            Received event.

        See Also
        --------
        Match
        """
        assert self._midi_in
        event = None
        polling = True
        elapsed = 0 if timeout else -1
        timeout = 0 if not timeout else timeout
        timeout = 0 if timeout < 0 else timeout
        while polling and timeout > elapsed:
            try:
                raw_message = self._midi_in.get_message()
                event = MidiEvent(*raw_message) if raw_message else None
                logger.debug(f'MIDI event: {event}')
                if event and not match:
                    polling = False
                elif event and isinstance(match, list) and match == event.message:  # noqa
                    polling = False
                elif event and isinstance(match, Match) and match.contains(event.message):  # noqa
                    polling = False
                time.sleep(read_delay)
                elapsed += .1 if timeout and timeout > 0 else 0
            except KeyboardInterrupt:
                logger.debug('\nPolling terminated.')
                raise
        return event

    def clear_event_queue(self, *, read_delay=.001):
        while self._midi_in and self._midi_in.get_message():
            time.sleep(read_delay)


class Interface:
    """Interface of the launchpad.
    """

    DAW = 'daw'
    MIDI = 'midi'
    READBACK_POSITION = 7

    class MidiWord(enum.IntEnum):
        DAW = 0x00
        MIDI = 0x01

    def __init__(self, midi_event):
        if len(midi_event.message) != MIDI_MESSAGE_LENGTH:
            raise RuntimeError('Unexpected MIDI message length; '
                               f'expected {MIDI_MESSAGE_LENGTH}, '
                               f'got {len(midi_event.message)}.')
        self._midi_event = midi_event
        midi_value = midi_event.message[Interface.READBACK_POSITION]
        self._interface = (Interface.MIDI
                           if midi_value == Interface.MidiWord.MIDI
                           else Interface.DAW)

    def __eq__(self, other):
        if (not isinstance(other, str)
                or (other.lower() != Interface.MIDI
                    and other.lower() != Interface.DAW)):
            return False
        return self.midi_event == other.midi_event

    def __repr__(self):
        return ('Interface()'
                if not self._interface
                else f"Interface('Interface.{self._interface.upper()}')")

    @property
    def midi_event(self):
        """MIDI event.
        """
        return self._midi_event


class Mode:
    """A Launchpad mode.
    """

    LIVE = 'live'
    PROG = 'prog'
    READBACK_POSITION = 7

    class MidiWord(enum.IntEnum):
        LIVE = 0x00
        PROG = 0x01

    def __init__(self, midi_event):
        if len(midi_event.message) != MIDI_MESSAGE_LENGTH:
            raise RuntimeError('Unexpected MIDI message length; '
                               f'expected {MIDI_MESSAGE_LENGTH}, '
                               f'got {midi_event.message}.')
        self._midi_event = midi_event
        midi_value = midi_event.message[Mode.READBACK_POSITION]
        self._mode = (Mode.LIVE
                      if midi_value == Mode.MidiWord.LIVE
                      else Mode.PROG)

    def __eq__(self, other):
        if (not isinstance(other, str)
                or (other.lower() != Mode.LIVE
                    and other.lower() != Mode.PROG)):
            return False
        return self.midi_event == other.midi_event

    def __repr__(self):
        return ('Mode()'
                if not self._mode
                else f"Mode('Mode.{self._mode.upper()}')")

    @property
    def midi_event(self):
        """MIDI event.
        """
        return self._midi_event


class Layout:
    """A Launchpad layout.
    """

    SESSION = 'session'
    CUSTOM_1 = 'custom_1'
    CUSTOM_2 = 'custom_2'
    CUSTOM_3 = 'custom_3'
    DAW_FADERS = 'daw_faders'
    PROG = 'prog'
    READBACK_POSITION = 7

    class MidiWord(enum.IntEnum):
        SESSION = 0x00
        CUSTOM_1 = 0x04
        CUSTOM_2 = 0x05
        CUSTOM_3 = 0x06
        DAW_FADERS = 0x0d
        PROG = 0x7f

    def __init__(self, midi_event):
        if len(midi_event.message) != MIDI_MESSAGE_LENGTH:
            raise RuntimeError('Unexpected MIDI message length; '
                               f'expected {MIDI_MESSAGE_LENGTH}, '
                               f'got {len(midi_event.message)}.')
        self._midi_event = midi_event
        midi_value = midi_event.message[Layout.READBACK_POSITION]
        self._layout = self._determine_layout(midi_value)

    def __eq__(self, other):
        if (not isinstance(other, str)
                or (other.lower() != Layout.SESSION
                    and other.lower() != Layout.CUSTOM_1
                    and other.lower() != Layout.CUSTOM_2
                    and other.lower() != Layout.CUSTOM_3
                    and other.lower() != Layout.DAW_FADERS
                    and other.lower() != Layout.PROG)):
            return False
        return self.midi_event == other.midi_event

    def __repr__(self):
        return ('Layout()'
                if not self._layout
                else f"Layout('Layout.{self._layout.upper()}')")

    @property
    def midi_event(self):
        """Midi event.
        """
        return self._midi_event

    def _determine_layout(self, midi_value):
        if midi_value == Layout.MidiWord.SESSION:
            layout = Layout.SESSION
        elif midi_value == Layout.MidiWord.CUSTOM_1:
            layout = Layout.CUSTOM_1
        elif midi_value == Layout.MidiWord.CUSTOM_2:
            layout = Layout.CUSTOM_2
        elif midi_value == Layout.MidiWord.CUSTOM_3:
            layout = Layout.CUSTOM_3
        elif midi_value == Layout.MidiWord.DAW_FADERS:
            layout = Layout.DAW_FADERS
        else:
            layout = Layout.PROG
        return layout


class MidiClient:
    """A MIDI client.
    """

    def __init__(self, client_name, client_number):
        self._client_name = client_name
        self._client_number = client_number
        self._out_ports = []
        self._in_ports = []

    def __eq__(self, other):
        if not isinstance(other, MidiClient):
            return False
        return (self.client_name == other.client_name
                and self.client_number == other.client_number)

    def __repr__(self):
        return ("MidiClient("
                f"name='{self.client_name}', "
                f"number={self.client_number})")

    @property
    def daw_out_port(self):
        """DAW out port.
        """
        daw_ports = list(filter(lambda port: re.search(r'1|DA',
                                port.port_name),
                                self._out_ports))
        return daw_ports[0] if len(daw_ports) > 0 else None

    @property
    def daw_in_port(self):
        """DAW in port.
        """
        daw_ports = list(filter(lambda port: re.search(r'1|DA',
                                port.port_name),
                                self._in_ports))
        return daw_ports[0] if len(daw_ports) > 0 else None

    @property
    def midi_out_port(self):
        """MIDI out port.
        """
        midi_ports = list(filter(lambda port: re.search(r'2|MI',
                                 port.port_name),
                                 self._out_ports))
        return midi_ports[0] if len(midi_ports) > 0 else None

    @property
    def midi_in_port(self):
        """MIDI in port.
        """
        midi_ports = list(filter(lambda port: re.search(r'2|MI',
                                 port.port_name),
                                 self._in_ports))
        return midi_ports[0] if len(midi_ports) > 0 else None

    @property
    def ports(self):
        """Ports.
        """
        return self._out_ports + self._in_ports

    @property
    def out_ports(self):
        """Out ports.
        """
        return self._out_ports

    @property
    def in_ports(self):
        """In ports.
        """
        return self._in_ports

    @property
    def client_name(self):
        """Client name.
        """
        return self._client_name

    @property
    def client_number(self):
        """Client number.
        """
        return self._client_number

    def is_open(self):
        """Checks if ports all ports are open.

        Returns
        -------
        bool
            `True` if all ports are open, otherwise `False`.
        """
        return all(port.is_open() for port in self.ports)

    def open(self, interface=Interface.MIDI):
        """Opens a MIDI port for interface `interface`.

        Parameters
        ----------
        interface : str, optional
            Interface to open.

        See Also
        --------
        Interface
        """
        self.close()
        if interface == Interface.DAW:
            self.daw_in_port.open()
            self.daw_out_port.open()
        elif interface == Interface.MIDI:
            self.midi_in_port.open()
            self.midi_out_port.open()
        else:
            raise ValueError('Must be a valid Interface.')

    def close(self):
        """Closes MIDI ports.
        """
        for port in self.ports:
            if port.is_open():
                port.close()

    def append_out_port(self, port):
        """Appends out port `port`.

        Parameters
        ----------
        port : MidiPort
            Port to append.

        See Also
        --------
        MidiPort
        """
        if not isinstance(port, MidiPort):
            raise TypeError('Must be of type MidiPort.')
        if port not in self._out_ports:
            self._out_ports.append(port)

    def append_in_port(self, port):
        """Appends in port `port`.

        Parameters
        ----------
        port : MidiPort
            Port to append.

        See Also
        --------
        MidiPort
        """
        if not isinstance(port, MidiPort):
            raise TypeError('Must be of type MidiPort.')
        if port not in self._in_ports:
            self._in_ports.append(port)


class SystemMidiPortParser:
    """System-specific way of parsing MIDI port names.
    """

    def __init__(self, in_ports, out_ports):
        lp_in_ports = list(filter(lambda port: re.search(r'Mini\s*MK3',
                                                         port,
                                                         re.IGNORECASE),
                                  in_ports))
        lp_out_ports = list(filter(lambda port: re.search(r'Mini\s*MK3',
                                                          port,
                                                          re.IGNORECASE),
                                   out_ports))
        self._found_clients = []
        self._parse(lp_in_ports,
                    in_ports,
                    MidiPort.IN)
        self._parse(lp_out_ports,
                    out_ports,
                    MidiPort.OUT)

    def _parse(self, lp_ports, ports, direction):
        ClientData = namedtuple('ClientData',
                                ['client_name',
                                 'client_number',
                                 'ports'])
        PortData = namedtuple('PortData',
                              ['port_name',
                               'port_number',
                               'port_index',
                               'system_port_name',
                               'direction'])
        for index, system_port_name in enumerate(lp_ports):
            linux_match = re.match(r'(.+):(.+)\s(\d+):(\d+)$', system_port_name)  # noqa
            windows_match = re.match(r'^(.+)\s\d+$', system_port_name)
            windows_match2 = re.match(r'(.+)\s\((.+)\)\s\d+$', system_port_name)  # noqa
            mac_match = re.match(r'(.+\sMK3)\s(.+)$', system_port_name)

            client_name = ''
            port_name = ''
            client_number = int(index + 1)
            port_number = int(index + 1)

            if linux_match:
                client_name = linux_match.group(1)
                port_name = linux_match.group(2)
                client_number = int(linux_match.group(3))
                port_number = int(linux_match.group(4))
            elif windows_match and not windows_match2:
                client_name = windows_match.group(1)
                port_name = system_port_name
            elif windows_match2:
                client_name = windows_match2.group(2)
                port_name = system_port_name
            elif mac_match:
                client_name = mac_match.group(1)
                port_name = system_port_name

            port_index = ports.index(system_port_name)
            port_data = PortData(port_name,
                                 port_number,
                                 port_index,
                                 system_port_name,
                                 direction)
            existing_client_data = list(filter(lambda c: c.client_name == client_name,  # noqa
                                        self._found_clients))
            existing_client_data = (existing_client_data[0]
                                    if len(existing_client_data) > 0
                                    else None)
            if not existing_client_data:
                client_data = ClientData(client_name,
                                         client_number,
                                         [])
                client_data.ports.append(port_data)
                self._found_clients.append(client_data)
            else:
                existing_client_data.ports.append(port_data)

    @property
    def found_clients(self):
        """Found clients.
        """
        return self._found_clients



"""Software model and discovery functions for Launchpad Mini MK3.
"""
from rtmidi import MidiOut, MidiIn
from .components import Grid, Panel
from .utils import SystemMidiPortParser, \
                   MidiPort, MidiClient, \
                   Interface, Mode, Layout # noqa
from .midi_messages import SysExMessages


class LaunchpadMiniMk3:
    """A Launchpad Mini MK3 device.
    """

    def __init__(self, midi_client):
        self._midi_client = midi_client

    def __eq__(self, other):
        if not isinstance(other, LaunchpadMiniMk3):
            return False
        return self.id == other.id

    def __repr__(self):
        return f'LaunchpadMiniMk3(id={self.id})'

    def is_open(self):
        """Returns `True` if the device is open, otherwise
        returns `False`. An open device is a device that can
        be read from and/or written to.

        Returns
        -------
        Boolean
            `True` if device is open, otherwise `False`.
        """
        return self._midi_client.is_open()

    def open(self, interface=Interface.MIDI):
        """Opens interface `interface` on this device for reading
        and writing.

        Parameters
        ----------
        interface : str, optional
            Interface to open.

        Raises
        ------
        ValueError
            If `interface` is invalid.

        See Also
        --------
        Interface
        """
        self._midi_client.open(interface)
        self.interface = interface

    def close(self):
        """Closes this device.
        """
        self._midi_client.close()

    def send_message(self, msg, *, interface=Interface.MIDI):
        """Sends a MIDI message `msg` to the `interface` of
        this device.

        Parameters
        ----------
        message : list
            MIDI message.
        interface : str, optional
            Interface to which to send message.

        Raises
        ------
        ValueError
            If `interface` is invalid.
        RuntimeError
            If device is closed.
        """
        if self.is_open():
            if interface == Interface.DAW:
                self.daw_out_port.send_message(msg)
            elif interface == Interface.MIDI:
                self.midi_out_port.send_message(msg)
            else:
                raise ValueError('Must be a valid Interface')
        else:
            raise RuntimeError('Port closed.')

    def poll_for_event(self, *, interface=Interface.MIDI,
                       timeout=5, match=None):
        """Polls for a MIDI event from interface `interface` either until
        `timeout` is reached or `match` is found.

        Parameters
        ----------
        interface : str, optional
            Interface to poll.
        timeout : float, optional
            Duration in seconds to wait for event before returning `None`.
        match : Match or None, optional
            Match criterion.

        Returns
        -------
        MidiEvent
            MIDI event received.

        Raises
        ------
        ValueError
            If `interface` is invalid.
        RuntimeError
            If this device is closed.

        See Also
        --------
        Interface
        MidiEvent
        """
        if self.is_open():
            if interface == Interface.DAW:
                return self.daw_in_port.poll_for_event(timeout=timeout,
                                                       match=match)
            elif interface == Interface.MIDI:
                return self.midi_in_port.poll_for_event(timeout=timeout,
                                                        match=match)
            else:
                raise ValueError('Must be a valid Interface.')
        else:
            raise RuntimeError('Port closed.')

    def clear_event_queue(self, *, interface=Interface.MIDI):
        """Clears MIDI event queue for interface `interface`.

        Parameters
        ----------
        interface : str
            Interface to clear.

        Raises
        ------
        ValueError
            If `interface` is invalid.
        RuntimeError
            If device is closed.

        See Also
        --------
        Interface
        """
        if self.is_open():
            if interface == Interface.DAW:
                return self.daw_in_port.clear_event_queue()
            elif interface == Interface.MIDI:
                return self.midi_in_port.clear_event_queue()
            else:
                raise ValueError('Must be a valid Interface.')
        else:
            raise RuntimeError('Port closed.')

    @property
    def id(self):
        """Unique ID of this device.
        """
        return self._midi_client.client_number

    @property
    def daw_in_port(self):
        """DAW interface MIDI-in port.

        Used for internal purposes.
        """
        return self._midi_client.daw_in_port

    @property
    def daw_out_port(self):
        """DAW interface MIDI-out port.

        Used for internal purposes.
        """
        return self._midi_client.daw_out_port

    @property
    def midi_in_port(self):
        """MIDI interface MIDI-in port.

        Used for internal purposes.
        """
        return self._midi_client.midi_in_port

    @property
    def midi_out_port(self):
        """MIDI interface MIDI-out port.

        Used for internal purposes.
        """
        return self._midi_client.midi_out_port

    @property
    def interface(self):
        """Interface of Launchpad.

        See Also
        --------
        Interface
        """
        self.send_message(SysExMessages.Interfaces.READBACK)
        return Interface(self.poll_for_event())

    @interface.setter
    def interface(self, value):
        if value.lower() == Interface.MIDI:
            self.send_message(SysExMessages.Interfaces.MIDI)
        elif value.lower() == Interface.DAW:
            self.send_message(SysExMessages.Interfaces.DAW)
        else:
            raise ValueError('Invalid interface set.')

    @property
    def panel(self):
        """Panel of Launchpad; represents the 9x9 array of buttons,
        with face buttons included.

        See Also
        --------
        Panel
        """
        return Panel(self)

    @property
    def grid(self):
        """Grid of Launchpad; represents the 8x8 array of white,
        faceless buttons.

        See Also
        --------
        Grid
        """
        return Grid(self)

    @property
    def mode(self):
        """
        Mode of Launchpad.

        See Also
        --------
        Mode
        """
        self.send_message(SysExMessages.Modes.READBACK)
        return Mode(self.poll_for_event())

    @mode.setter
    def mode(self, value):
        if value.lower() == Mode.LIVE:
            self.send_message(SysExMessages.Modes.LIVE)
        elif value.lower() == Mode.PROG:
            self.send_message(SysExMessages.Modes.PROG)
        else:
            raise ValueError('Invalid mode set.')

    @property
    def layout(self):
        """Layout of Launchpad.

        See Also
        --------
        Layout
        """
        self.send_message(SysExMessages.Layouts.READBACK)
        return Layout(self.poll_for_event())

    @layout.setter
    def layout(self, value):
        if value.lower() == Layout.SESSION:
            self.send_message(SysExMessages.Layouts.SESSION)
        elif value.lower() == Layout.CUSTOM_1:
            self.send_message(SysExMessages.Layouts.CUSTOM_1)
        elif value.lower() == Layout.CUSTOM_2:
            self.send_message(SysExMessages.Layouts.CUSTOM_2)
        elif value.lower() == Layout.CUSTOM_3:
            self.send_message(SysExMessages.Layouts.CUSTOM_3)
        elif value.lower() == Layout.DAW_FADERS:
            self.send_message(SysExMessages.Layouts.DAW_FADERS)
        elif value.lower() == Layout.PROG:
            self.send_message(SysExMessages.Layouts.PROG)
        else:
            raise ValueError('Invalid layout set.')

    def device_inquiry(self):
        """Sends a "device inquiry" message to this device.

        Returns
        -------
        MidiEvent
            MIDI event received.
        """
        self.mode = Mode.PROG
        self.send_message(SysExMessages.DEVICE_INQUIRY)
        return self.poll_for_event()


def find_launchpads():
    """Searches for connected Launchpad Mini MK3 devices.

    Returns
    -------
    list of LaunchpadMiniMk3
        List of found devices.
    """
    midi_out = MidiOut()
    out_ports = midi_out.get_ports()
    midi_in = MidiIn()
    in_ports = midi_in.get_ports()
    parser = SystemMidiPortParser(in_ports, out_ports)

    found_launchpads = []
    for client_data in parser.found_clients:
        midi_client = MidiClient(client_data.client_name,
                                 client_data.client_number)

        for port_data in client_data.ports:
            midi_port = MidiPort(port_data.port_name,
                                 port_data.port_number,
                                 port_data.port_index,
                                 port_data.system_port_name,
                                 direction=(MidiPort.OUT
                                            if port_data.direction == MidiPort.OUT  # noqa
                                            else MidiPort.IN),
                                 midi_out=midi_out,
                                 midi_in=midi_in)
            if port_data.direction == MidiPort.OUT:
                midi_client.append_out_port(midi_port)
            elif port_data.direction == MidiPort.IN:
                midi_client.append_in_port(midi_port)

        found_launchpads.append(LaunchpadMiniMk3(midi_client))

    return found_launchpads
